/**************************************************************************
 * JumpDriveMFD V1.0 - FTL travel in a difficult fashion for O2K6
 *     created: July, 6th 2006
 * last update: September, 10th 2006
 * Copyright (c) 2006 Friedrich Kastner-Masilko, M.Sc.
 *
 * JumpDrive class header
 **************************************************************************/

// Trigonimetric definitions
#define TAN70			2.747477419
#define RAD90			1.570796327
#define RAD70			1.221730476

// Time definitions
#define ACCTIME			3.6
#define FLYTIME			2.1
#define JMPTIME			1.6
#define DECTIME			7.3

// Drive states
#define S_IDLE			0
#define S_LOAD			1
#define S_ACCELERATE	2
#define S_CRUISE		3
#define S_JUMP			4
#define S_DECELERATE	5

// Includes
#include "windows.h"
#include "orbitersdk.h"
#include "orbitersoundsdk3.h"
#include "JumpDrive.h"
#include <stdio.h>

// Constructor - Creates a drive with the given data
JumpDrive::JumpDrive(JumpDriveData data)
{
	this->vessel=data.vessel;
	this->sound=data.sound;
	this->env=data.env;
	this->loadtime=data.loadtime;
	this->capacity=data.capacity;
	this->offset=data.offset;
	this->jump=data.jump;
	this->speed=data.speed;
	this->cells=data.cells;
	filledCells=0;
	object=vessel->GetHandle();
	addvec=_V(0,1,0);	
	engaged=0;
	oldRVEL=_V(0,0,0);	
	target=_V(0,0,0);	
	state=S_IDLE;
	beacon.shape=BEACONSHAPE_STAR;
	beacon.pos=&beaconpos;
	beacon.col=&beaconcol;
	beacon.size=1.0;
	beacon.falloff=1;
	beacon.period=0;
	beacon.duration=0;
	beacon.tofs=0;
	beacon.active=true;
	beaconcol=_V(1,1,1);		
	reference=oapiGetGbodyByIndex(0);
	snd_load	=data.soundSlot[0];
	snd_wooshout=data.soundSlot[1];
	snd_bang	=data.soundSlot[2];
	snd_wooshin	=data.soundSlot[3];
	fuel=0;
	fuelSource=0;
}

// Destructor - removes the internal jump range table
JumpDrive::~JumpDrive()
{
	delete [] jump;
}

//Sets the current fuel source index
void JumpDrive::SetFuelSource(int index)
{	
	fuelSource=index;
}

//Returns the current fuel source index	
int JumpDrive::GetFuelSource()
{
	return fuelSource;
}

//Returns amount of available fuel cells
int JumpDrive::GetFuelCells()
{
	return cells;
}

//Returns amount of filled fuel cells
int JumpDrive::GetFilledCells()
{
	return filledCells;
}

//Returns internal fuel level (0-1)
double JumpDrive::GetFuel()
{
	return fuel;
}

//Returns current jump range in meters
double JumpDrive::GetJumpRange(int cells)
{
	return jump[cells];
}

//Returns the sound handle
int JumpDrive::GetSoundHandle()
{
	return sound;
}

//Starts a jump procedure if jump is possible.
bool JumpDrive::Start()
{
	if (isJumpPossible())
	{
		VECTOR3 vec;		
		vessel->GlobalRot(_V(0,0,speed), addvec);
		vessel->GlobalRot(_V(0,0,jump[filledCells]), target);
		engaged=oapiGetSimTime();
		if (oapiCameraTarget()==object)
		{
			env->FOV=oapiCameraAperture();
			env->Reset=false;
		}
		vessel->GetRelativeVel(reference, oldRVEL);
		vessel->GetRelativePos(reference, vec);
		target+=vec;		
		state=S_ACCELERATE;		
		PlayVesselWave3(sound,snd_wooshout,NOLOOP,255);
		fuel=0.99;
		filledCells=0;
		return true;
	}
	return false;
}

//Starts a loading procedure
bool JumpDrive::Load()
{
	PROPELLANT_HANDLE propellant=vessel->GetPropellantHandleByIndex(fuelSource);
	if (propellant!=NULL)		
	{
		double mass=vessel->GetPropellantMass(propellant);
		if (mass>0.0 && (fuel<=0.0 || (fuel>=1.0 && filledCells+1<cells)))
		{
			if (fuel>=1.0) filledCells++;
			fuel=0.0;
			state=S_LOAD;			
			return true;
		}
	}
	return false;
}

//Engages a fleet command to the current drive's owner with the commanding
//vessel in parameter 1 and the command range in meters in parameter 2
void JumpDrive::FleetCommand(VESSEL *vessel, double treshold)
{
	if (isJumpPossible())
	{
		VECTOR3 d;
		oapiGetRelativePos (this->object, vessel->GetHandle(), &d);
		if (length(d)<treshold)
		{
			vessel->GetGlobalOrientation(d);
			this->vessel->SetGlobalOrientation(d);
			Start();
		}
	}
}

//Returns wether or not the vessel is currently jumping
bool JumpDrive::Jumping()
{
	return state>S_LOAD;
}

//Returns wether or not a jump is possible
bool JumpDrive::isJumpPossible()
{
	if (fuel<1.0) return false;
	return (vessel->GetFlightStatus() & 0x02)==0;/*Alternative implementation:
	int i=0;
	DOCKHANDLE dock;
	while(true)
	{
		dock=vessel->GetDockHandle(i++);
		if (dock==NULL) return true;
		if (vessel->GetDockStatus(dock)!=NULL) return false;
	}*/
}

//Delete all docked vessels from the parent recursively
void JumpDrive::deleteVessels(VESSEL *vessel, OBJHANDLE parent)
{
	int i=0;
	DOCKHANDLE dock;
	OBJHANDLE dockHandle;
	VESSEL *child;	
	while(true)
	{
		dock=vessel->GetDockHandle(i++);
		if (dock==NULL) break;
		if ((dockHandle=vessel->GetDockStatus(dock))!=NULL)
		{
			if (dockHandle!=parent)
			{
				deleteVessels((child=new VESSEL(dockHandle, 1)), vessel->GetHandle());
				delete child;
				oapiDeleteVessel(dockHandle);
			}
		}
	}	
}

//Processing jump drive progress - should be called periodically
void JumpDrive::Process(double SimT,double SimDT)
{	
	OBJHANDLE cam=oapiCameraTarget();
	double diff;
	//Loading a cell
	if (state==S_LOAD)
	{
		PROPELLANT_HANDLE propellant=vessel->GetPropellantHandleByIndex(fuelSource);
		if (propellant!=NULL)		
		{
			double mass=vessel->GetPropellantMass(propellant);
			if (mass>0)
			{
				diff=SimDT/loadtime;
				if (diff>(1.0-fuel))
				{
					diff=1.0-fuel;					
					state=S_IDLE;					
				}						
				if (diff>mass/capacity) diff=mass/capacity;
				fuel+=diff;
				mass-=diff*capacity;
				vessel->SetPropellantMass(propellant, mass);
				if (state!=S_LOAD) StopVesselWave3(sound, snd_load);
				else PlayVesselWave3(sound,snd_load,LOOP,255);
			}
			else StopVesselWave3(sound, snd_load);
		}
		else StopVesselWave3(sound, snd_load);
	}
	if (state>S_LOAD)
	{
		diff=SimT-engaged;
		VESSELSTATUS2 status;
		status.version=2;
		status.flag=0;		
		double temp=(RAD90+RAD70)/2;
		switch(state)
		{
		//Accelerating to jump speed
		case S_ACCELERATE:
			if (diff>ACCTIME)
			{				
				engaged=SimT;				
				state=S_CRUISE;				
			}
			else
			{	
				vessel->GetStatusEx(&status);
				temp=diff/ACCTIME;
				status.rvel+=addvec*SimDT*temp;				
				vessel->DefSetStateEx(&status);
				fuel=0.99-temp*0.2;
				temp=(RAD90+temp*RAD70)/2;
			}
			break;
		//Cruise at jump speed
		case S_CRUISE:
			if (diff>FLYTIME)
			{
				vessel->GetStatusEx(&status);
				hyperobjectIn=oapiCreateVesselEx("Hyperspace","Hyperspace",&status);
				vessel->GetRelativeVel(reference, status.rvel);
				status.rpos=target;
				status.rbody=reference;				
				vessel->DefSetStateEx(&status);
				PlayVesselWave3(sound,snd_bang,NOLOOP,255);
				engaged=SimT;				
				state=S_JUMP;				
				hyperobjectOut=oapiCreateVesselEx("Hyperspace","Hyperspace",&status);
				hyperspaceIn=oapiGetVesselInterface(hyperobjectIn);
				hyperspaceOut=oapiGetVesselInterface(hyperobjectOut);
				hyperspaceIn->SetVisibilityLimit(1e-12);
				hyperspaceOut->SetVisibilityLimit(1e-12);
				vessel->GetCameraOffset(beaconpos);	
				beaconpos.z+=offset;
				hyperspaceIn->AddBeacon(&beacon);
				hyperspaceOut->AddBeacon(&beacon);
				//Delete all docked vessels (hyperspace physics are dangerous, you know...)
				deleteVessels(vessel, NULL);
			}
			else
			{
				
			}
			break;
		//Explosion effect
		case S_JUMP:
			if (diff>JMPTIME)
			{
				PlayVesselWave3(sound,snd_wooshin,NOLOOP,255);
				engaged=SimT;				
				state=S_DECELERATE;
				oapiDeleteVessel(hyperobjectIn);
				oapiDeleteVessel(hyperobjectOut);
				fuel=0.2;
			}
			else
			{
				if (diff>0.3) diff=0;
				else diff/=0.3;
				if (diff!=0) diff=1/diff/diff;				
				beacon.size=diff*10;				
			}
			break;
		//Decelerating from jump speed
		case S_DECELERATE:
			if (diff>DECTIME)
			{
				vessel->GetStatusEx(&status);
				vessel->GetRelativeVel(reference, status.rvel);
				vessel->GetRelativePos(reference, status.rpos);
				status.rbody=reference;
				status.rvel=oldRVEL;
				vessel->DefSetStateEx(&status);
				temp=env->FOV;
				state=S_IDLE;				
				fuel=0;
			}
			else
			{
				vessel->GetStatusEx(&status);
				temp=(DECTIME-diff)/DECTIME/2;
				status.rvel-=addvec*SimDT*temp;				
				vessel->DefSetStateEx(&status);
				fuel=temp*0.4;
				temp=(RAD90+temp*RAD70*2)/2;
			}
			break;
		}
		//Setting FOV if jumping vessel has focus
		if (cam==object)
		{
			oapiCameraSetAperture(temp);
			env->Reset=false;
		}
	}	
	else if (cam==object && !env->Reset)
	{
		//Reseting FOV to environment setting if no jumping vessel has focus
		oapiCameraSetAperture(env->FOV);
		env->Reset=true;
	}
}